var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
import { saveContext, safeMixins, hasDimsProp, needRendering, } from '../util';
import { composeMatrix2X3, transformPoint2D, invertMatrix2X3, } from 'web-util-kit';
import { Control } from '../auxiliary/control';
import { ShapeEvent } from '../event';
export class ShapeBase extends ShapeEvent {
    constructor(options) {
        super();
        this.type = 'shape-base';
        // dimensions of the shape
        this.left = 0;
        this.top = 0;
        this._width = 0;
        this._height = 0;
        // center point relative ratio
        this._centerXRatio = 0.5;
        this._centerYRatio = 0.5;
        // drawing style
        this.lineWidth = 2;
        this.strokeWidth = 2;
        this.fillStyle = 'blue';
        this.strokeStyle = 'yellow';
        this.lineDash = [];
        // more style put here
        // transform
        // 角度单位
        this.angle = 0;
        this.skewX = 0;
        this.skewY = 0;
        this._scaleX = 1;
        this._scaleY = 1;
        this.flipX = false;
        this.flipY = false;
        this.path2D = new Path2D();
        this.active = false;
        // controll points 
        this.control = new Control();
        // extra info stored in shape for other usage
        this.other = {};
        this.vinci = null;
        safeMixins(this, options);
    }
    set width(nextVal) {
        if (nextVal < 0) {
            return;
        }
        this._width = nextVal;
    }
    get width() { return this._width; }
    set height(nextVal) {
        if (nextVal < 0) {
            return;
        }
        this._height = nextVal;
    }
    get height() { return this._height; }
    set centerXRatio(nextVal) {
        if (nextVal < 0 || nextVal > 1) {
            return;
        }
        this._centerXRatio = nextVal;
    }
    get centerXRatio() { return this._centerXRatio; }
    set centerYRatio(nextVal) {
        if (nextVal < 0 || nextVal > 1) {
            return;
        }
        this._centerYRatio = nextVal;
    }
    get centerYRatio() { return this._centerYRatio; }
    set scaleX(val) {
        if (val < 0) {
            return;
        }
        this._scaleX = val;
    }
    get scaleX() {
        return this._scaleX;
    }
    set scaleY(val) {
        if (val < 0) {
            return;
        }
        this._scaleY = val;
    }
    get scaleY() {
        return this._scaleY;
    }
    get translateX() {
        return this.left + this.width * this.centerXRatio;
    }
    get translateY() {
        return this.top + this.height * this.centerYRatio;
    }
    set(arg1, arg2) {
        let nextOpts = {};
        if (typeof arg1 === 'string' && arg2 !== undefined) {
            nextOpts = {
                [arg1]: arg2
            };
        }
        else if (arg1 && typeof arg1 === 'object') {
            nextOpts = { ...arg1 };
        }
        safeMixins(this, nextOpts);
        this.normalize(nextOpts);
    }
    setVinci(vinci) {
        this.vinci = vinci;
        this.dispose();
        const renderer = () => vinci.requestRender();
        this.onAdded(renderer);
        this.onRemoved(renderer);
        this.onSelected(renderer);
        this.onUnSelected(renderer);
        this.onModified(renderer);
    }
    getTransformOptions() {
        return {
            angle: this.angle,
            skewX: this.skewX,
            skewY: this.skewY,
            scaleX: this.scaleX,
            scaleY: this.scaleY,
            flipX: this.flipX,
            flipY: this.flipY,
            translateX: this.translateX,
            translateY: this.translateY,
        };
    }
    getTransformMatrix2X3Array() {
        return composeMatrix2X3({
            angle: this.angle,
            skewX: this.skewX,
            skewY: this.skewY,
            scaleX: this.scaleX,
            scaleY: this.scaleY,
            flipX: this.flipX,
            flipY: this.flipY,
            translateX: this.translateX,
            translateY: this.translateY,
        });
    }
    normalize(nextOptions) {
        if (hasDimsProp(nextOptions)) {
            this.calcShape();
        }
        this.calcDimensions();
    }
    calcControls() {
        this.control.calculate(this.width, this.height);
    }
    renderControls(ctx, vpt) {
        if (!this.active) {
            return;
        }
        this.calcControls();
        // TODO check is control visible
        ctx.strokeStyle = 'blue';
        ctx.setLineDash([4, 4]);
        ctx.strokeRect(this.left - this.translateX, this.top - this.translateY, this.width, this.height);
        ctx.beginPath();
        const mt = this.control.getPoint('mt');
        const mtr = this.control.getPoint('mtr');
        ctx.moveTo(mt.x, mt.y);
        ctx.lineTo(mtr.x, mtr.y);
        ctx.stroke();
        ctx.fillStyle = 'red';
        this.control.getAllVisiblePoints().forEach((cp) => {
            const point = cp.point;
            ctx.beginPath();
            ctx.arc(point.x, point.y, 6, 0, Math.PI * 2);
            ctx.fill();
        });
    }
    isPointInPath(ctx, point) {
        if (ctx.isPointInPath(this.path2D, point.x, point.y)) {
            return true;
        }
        return false;
    }
    rotate(angle) {
        this.angle = angle;
    }
    move(x, y) {
        this.left = x;
        this.top = y;
        this.calcShape();
    }
    scaleToControlPoint(pointType, pointer, lastPointer, lastDims) {
        const inverseMatrix = invertMatrix2X3(this.getTransformMatrix2X3Array());
        pointer = transformPoint2D(pointer, inverseMatrix);
        lastPointer = transformPoint2D(lastPointer, inverseMatrix);
        const diff = pointer.sub(lastPointer);
        if (pointType === 'tl') {
            this.left = pointer.x + this.translateX;
            this.top = pointer.y + this.translateY;
            this.width = lastDims.width - diff.x;
            this.height = lastDims.height - diff.y;
        }
        else if (pointType === 'tr') {
            this.left = pointer.x + this.translateX - this.width;
            this.top = pointer.y + this.translateY;
            this.width = lastDims.width + diff.x;
            this.height = lastDims.height - diff.y;
        }
        else if (pointType === 'bl') {
            this.left = pointer.x + this.translateX;
            this.top = pointer.y + this.translateY - this.height;
            this.width = lastDims.width - diff.x;
            this.height = lastDims.height + diff.y;
        }
        else if (pointType === 'br') {
            this.left = pointer.x + this.translateX - this.width;
            this.top = pointer.y + this.translateY - this.height;
            this.width = lastDims.width + diff.x;
            this.height = lastDims.height + diff.y;
        }
        else if (pointType === 'ml') {
            this.left = pointer.x + this.translateX;
            this.width = lastDims.width - diff.x;
        }
        else if (pointType === 'mr') {
            this.width = lastDims.width + diff.x;
        }
        else if (pointType === 'mt') {
            this.top = pointer.y + this.translateY;
            this.height = lastDims.height - diff.y;
        }
        else if (pointType === 'mb') {
            this.height = lastDims.height + diff.y;
        }
        this.calcShape();
    }
    findControlCorner(pointer) {
        return this.control.findType(pointer, this.getTransformMatrix2X3Array());
    }
    // TODO: layer index operation
    upward() {
        if (!this.vinci) {
            return;
        }
        this.vinci.upwardObject(this);
    }
    downward() {
        if (!this.vinci) {
            return;
        }
        this.vinci.downwardObject(this);
    }
    toFront() {
        if (!this.vinci) {
            return;
        }
        this.vinci.toFrontObject(this);
    }
    toBack() {
        if (!this.vinci) {
            return;
        }
        this.vinci.toBackObject(this);
    }
}
__decorate([
    needRendering()
], ShapeBase.prototype, "set", null);
__decorate([
    saveContext()
], ShapeBase.prototype, "renderControls", null);
__decorate([
    saveContext()
], ShapeBase.prototype, "isPointInPath", null);
__decorate([
    needRendering()
], ShapeBase.prototype, "rotate", null);
__decorate([
    needRendering()
], ShapeBase.prototype, "move", null);
__decorate([
    needRendering()
], ShapeBase.prototype, "scaleToControlPoint", null);
