import { decorators, objects } from 'util-kit';
import { DIMS_PROPERTIES, RENDERING_PROPERTIES, } from './const';
export * from './const';
const { createDecorator } = decorators;
const { deepClone, equals } = objects;
export function addListener(element, eventName, handler, options) {
    element && element.addEventListener(eventName, handler, options);
}
;
export function removeListener(element, eventName, handler, options) {
    element && element.removeEventListener(eventName, handler, options);
}
;
// export function isMainEvent(evt: Event) {
//  if (evt.isPrimary === true) {
//    return true;
//  }
//  if (evt.isPrimary === false) {
//    return false;
//  }
//  if (evt.type === 'touchend' && evt.touches.length === 0) {
//    return true;
//  }
//  if (evt.changedTouches) {
//    return evt.changedTouches[0].identifier === this.mainTouchId;
//  }
//  return true;
// }
export function saveContext() {
    return createDecorator((fn, key) => {
        return function (ctx, vpt) {
            if (!(ctx instanceof CanvasRenderingContext2D)) {
                throw new Error('failed to get CanvasRenderingContext2D');
            }
            this.normalize();
            ctx.save();
            ctx.lineWidth = this.lineWidth;
            ctx.fillStyle = this.fillStyle;
            ctx.strokeStyle = this.strokeStyle;
            ctx.setLineDash(this.lineDash);
            ctx.setTransform(...this.getTransformMatrix2X3Array());
            const res = fn.apply(this, arguments);
            ctx.restore();
            return res;
        };
    });
}
export function needRendering() {
    return createDecorator((fn, key) => {
        return function () {
            const prev = getRenderingProperties(this);
            const res = fn.apply(this, arguments);
            const current = getRenderingProperties(this);
            if (!equals(current, prev)) {
                this.fire('onModified');
            }
            return res;
        };
    });
}
export function hasDimsProp(obj) {
    if (!obj) {
        return false;
    }
    for (let i = 0; i < DIMS_PROPERTIES.length; i++) {
        const prop = DIMS_PROPERTIES[i];
        if (obj[prop] && typeof obj[prop] === 'number') {
            return true;
        }
    }
    return false;
}
export function getRenderingProperties(obj) {
    const res = {};
    RENDERING_PROPERTIES.forEach((prop) => {
        res[prop] = deepClone(obj[prop]);
    });
    return res;
}
export function safeMixins(target, source, backup = 'other') {
    if (!source || !target) {
        return;
    }
    target[backup] = target[backup] || {};
    const keys = Object.keys(source);
    keys.forEach((key) => {
        if (typeof target[key] === 'undefined') {
            target[backup][key] = source[key];
        }
        else {
            target[key] = source[key];
        }
    });
    const otherKeys = Object.keys(target[backup]);
    otherKeys.forEach((otherKey) => {
        if (typeof target[otherKey] !== 'undefined') {
            delete target[backup][otherKey];
        }
    });
}
